--[[
This file is part of Twinnings - a puzzle game

Twinnings is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
]]


require('saves')
require('chars')

bit = require('bit')


function getbit(num, n)
	return bit.rshift(num, n) % 2
end


function square(x, y, side, color)
	if color then love.graphics.setColor(color) end
	love.graphics.rectangle('fill', x, y, side, side)
end


-- custom run to prevent redraws when idle
function love.run()
	love.load(love.arg.parseGameArguments(arg), arg)
	-- We don't want the first frame's dt to include time taken by love.load.
	love.timer.step()
	local dt = 0
	-- Main loop time.
	return function()
		-- Process events.
		if love.event then
			love.event.pump()
			for name, a,b,c,d,e,f in love.event.poll() do
				if name == "quit" then
					if not love.quit or not love.quit() then
						return a or 0
					end
				end
				love.handlers[name](a,b,c,d,e,f)
			end
		end
		-- Update dt, as we'll be passing it to update
		dt = love.timer.step()
		-- Call update and draw
		if love.update then love.update(dt) end -- will pass 0 if love.timer is disabled
		if screen.changed and love.graphics.isActive() then
			love.graphics.origin()
			love.graphics.clear(love.graphics.getBackgroundColor())
			love.draw()
			love.graphics.present()
			screen.changed = false
		end
		love.timer.sleep(0.05)
	end
end


function love.load()
	VERSION = '0.4'
	C_SELECTION = {.9, .3, .3}
	C_PAIRED = {0,0,1, .5}
	C_BACKGROUND = {.3, .3, .3}
	last_rotation_time = nil
	if not restore_game() then
		restart_game()
		init(1)
		love.resize()
	end
end


function end_game(screen)
	local duration = os.time() - game.start
	local h = math.floor(duration / 3600)
	local m = math.floor(duration / 60) - h * 60 .. 'm '
	local s = duration % 60 .. 's'
	if h > 0 then h = h .. 'h ' else h = '' end
	del_game()
	restart_game()
	title_screen(screen, 'Time: ' .. h .. m .. s)
end


function restart_game()
	game = {
		seed = os.time(),
		retries = 4,
	}
	math.randomseed(game.seed)
end


function title_screen(m, osd)
	local messages = {
		'   TWI   NNI   NGSNEW   RESUMEQUIT  ',
		'c PAJO  2022  GPL3NEW   RESUMEQUIT  ',
		' CONGR ATULA TIONS      RETRY QUIT  ',
		' GAME  OVER             RETRY QUIT  ',
	}
	init(0, messages[m])
	current.page = m
	current.osd = osd or VERSION
end


function init(level, message)
	if level == 0 then
		game.root = 5
		game.salt = 0
		game.couples = 18
	elseif level == 1 then
		game.start = os.time()
		game.root = 2
		game.salt = 0
		game.couples = 2
	elseif level == 2 then
		game.root = 3
	elseif level == 4 then
		game.salt = 1
	elseif level == 6 then
		game.salt = 1
		game.couples = 8
	elseif level == 8 then
		game.salt = 2
	elseif level == 10 then
		game.couples = 18
	elseif level == 12 then
		game.root = 4
		game.salt = 5
		game.couples = 8
	elseif level == 13 then
		game.salt = 6
	elseif level == 14 then
		game.salt = 7
	elseif level == 16 then
		game.salt = 8
		game.couples = 18
	elseif level == 17 then
		end_game(3)
		return
	end
	game.bits = game.root ^ 2
	game.max_n = 2 ^ game.bits - 1 -- 0b111111111
	current = {
		combos = {},
		paired = {},
		selected = false,
		level = level,
		saved_combos = {},
		page = false,
	}
	if message then
		current.combos = spell(message)
	else
		generate_combos()
	end
	love.resize()
end


function love.resize()
	screen = {
		divide = function(self)
			self.slots = { side = self.side / self.root }
			for y = 0, self.root - 1 do for x = 0, self.root - 1 do
				self.slots[#self.slots + 1] = {
					x = self.x + x * self.slots.side,
					y = self.y + y * self.slots.side
				}
			end end
			return slots
		end,
		side = math.min(love.graphics.getWidth(), love.graphics.getHeight()),
		root = math.ceil(math.sqrt(game.couples * 2)),
		changed = true
	}
 	screen.x = (love.graphics.getWidth() - screen.side) / 2
 	screen.y = (love.graphics.getHeight() - screen.side) / 2
	screen:divide()
	screen.slots.margin = screen.slots.side * .1
end


function generate_combos()
	local function insert_at_random(t, v)
		local new_place = math.random(1, #t + 1)
		table.insert(t, new_place, v)
		return new_place
	end
	current.combos = {}
	for a = 1, game.couples do
		-- avoid creating pieces with all bits on or off
		local original = insert_at_random(current.combos, math.random(1, game.max_n - 1))
		-- create complementary number
		-- risky to use bit.bnot here, as it creates extra 111s out of shape range
		insert_at_random(current.combos, bit.bxor(current.combos[original], game.max_n))
	end
	-- salt
	for b = 1, game.salt do
		for a = 1, #current.combos do
			current.combos[a] = bit.bor(current.combos[a], 2 ^ math.random(0, game.bits - 1))
		end
	end
end


function draw_osd()
	local osd
	if current.osd then
		osd = current.osd
	else
		osd = 'Level ' .. current.level .. ' '
		for r = 1, game.retries do
			osd = osd .. '•'
		end
	end
	love.graphics.setColor(0, 0, 0)
	love.graphics.print(osd, 5, 5)
	love.graphics.setColor(1, 1, 1)
	love.graphics.print(osd, 4, 4)
end


function love.update(dt)
	if dt < 1/30 then
		love.timer.sleep(1/30 - dt)
	end
	if last_rotation_time and (love.timer.getTime() - last_rotation_time  > 1) then
		screen.changed = true
		last_rotation_time = nil
	end
end


function draw_small_rectangle(combo, xx, yy, small, color)
	local side = game.root
	local small_side = (screen.slots.side - 2 * screen.slots.margin) / side
	local s = small and .75 * small_side or 0
	for y = 0, side - 1 do for x = 0, side - 1 do
		if getbit(combo, y * side + x) == 1 then
			square(
				screen.slots.margin + xx + x * small_side + s / 2,
				screen.slots.margin + yy + y * small_side + s / 2,
				small_side - s,
				{1, 1, 1}
			)
		end
	end end
end


function draw_playfield()
	-- draw playfield background, except on special screens
	if current.level > 0 then
		square(screen.x, screen.y, screen.side, C_BACKGROUND)
	end
	-- draw individual combos
	for a = 1, #current.combos do
		-- mark pair of currently selected square
		if current.selected and current.selected == current.paired[a] then
			square(
				screen.slots[a].x,
				screen.slots[a].y,
				screen.slots.side,
				{0, 0, 0}
			)
		end
		-- mark currently selected square or square pointed by keyboard
		if current.cursor == a or current.selected == a then
			square(
				screen.slots[a].x,
				screen.slots[a].y,
				screen.slots.side,
				C_SELECTION
			)
		end
		-- inner square background
		square(
			screen.slots[a].x + screen.slots.margin,
			screen.slots[a].y + screen.slots.margin,
			screen.slots.side - 2 * screen.slots.margin,
			{0, 0, 0}
		)
		-- small and even smaller squares
		draw_small_rectangle(current.combos[a], screen.slots[a].x, screen.slots[a].y)
		if current.saved_combos[a] then
			draw_small_rectangle(current.saved_combos[a], screen.slots[a].x, screen.slots[a].y, true)
		end
		-- mark pairs
		if current.paired[a] then
			square(
				screen.slots[a].x,
				screen.slots[a].y,
				screen.slots.side,
				C_PAIRED
			)
		end
	end
end


function love.draw()
	draw_playfield()
	draw_osd()
end


function restore_game()
	saved_game = load_game()
	if not saved_game then return false end
	game, current = unpack(saved_game)
	love.resize()
	return true
end


function click_on_menu(c)
	if not c or current.combos[c] == 0 then
		return
	elseif c > 30 then
		love.event.push('quit')
		return
	end
	if current.page <= 2 then
		if c <= 18 then
			title_screen(current.page % 2 + 1)
		elseif c > 24 then
			restore_game()
		else
			init(1)
		end
		return
	elseif current.page > 2 and c > 24 then
		init(1)
		return
	end
end


function click_on_slot(c)
	local function unpair_slot(s)
		if current.paired[s] then
			current.paired[current.paired[s]] = false
			current.paired[s] = false
		end
	end
	local function check_victory()
		for a = 1, #current.combos do
			if not current.paired[a] then return false end
		end
		return true
	end
	if current.level == 0 then
		click_on_menu(c)
		return
	elseif c and current.combos[c] then
		local shape_a, shape_b = current.combos[c], current.combos[current.selected]
		if current.selected then
			-- on double click
			if c == current.selected then
				-- restore from saved if any
				if current.saved_combos[c] then
					current.combos[c] = current.saved_combos[c]
					current.saved_combos[c] = nil
					-- unpair, lose life
					unpair_slot(c)
					game.retries = game.retries - 1
					if game.retries < 0 then
						end_game(4)
					end
				end
				-- in any case, unselect shape
				current.selected = nil
			-- loose check
			elseif bit.bor(shape_a, shape_b) == game.max_n then
				-- strict check
				if bit.bxor(shape_a, shape_b) ~= game.max_n then
					-- save shape_a
					current.saved_combos[c] = current.saved_combos[c] or shape_a
					-- punch out shape_a
					current.combos[c] = bit.band(shape_a, bit.bnot(shape_b))
				end
				-- unmark old connections if any
				unpair_slot(c)
				unpair_slot(current.selected)
				-- reference the other slot
				current.paired[c] = current.selected
				current.paired[current.selected] = c
				current.selected = false
				if check_victory() then init(current.level + 1) end
			else
				current.selected = c
			end
		else
			current.selected = c
		end
	else
		-- outside click
		current.selected = false
	end
end


function love.mousepressed(x, y, button)
	screen.changed = true
	if button == 2 then
		if current.level > 0 then
			save_game({game, current})
			title_screen(1)
		else
			click_on_slot(25) -- resume
		end
		return
	end
	local function get_slot(x, y, domain)
		local function isin(xx, yy, x, y, w, h)
			return xx >= x and xx <= (x + w) and yy >= y and yy <= (y + h)
		end
		for a = 1, #domain.slots do
			if isin(x, y, domain.slots[a].x, domain.slots[a].y, domain.slots.side, domain.slots.side) then
				return a
			end
		end
	end
	local c = get_slot(x, y, screen)
	current.cursor = false
	click_on_slot(c)
end


function move_cursor(cursor, x, y, side)
	if not cursor then return 1 end
	local c = cursor
	if y == 0 then
		-- horizontal movement - stay in the same row
		c = cursor + x
		if c % side == .5 + .5 * x then
			c = c - side * x
		end
	else
		-- vertical movement - stay in the same column
		c = (c - 1 + y * side) % side^2 + 1
	end
	return c
end


function love.keypressed(k)
	screen.changed = true
	local root = screen.root
	if k == "q" then
		love.event.push("quit")
	elseif k == "escape" then
		if current.level == 0 then
			love.event.push("quit")
			return
		end
		save_game({game, current})
		title_screen(1)
	elseif k == "up" then
		current.cursor = move_cursor(current.cursor, 0, -1, root)
	elseif k == "down" then
		current.cursor = move_cursor(current.cursor, 0, 1, root)
	elseif k == "left" then
		current.cursor = move_cursor(current.cursor, -1, 0, root)
	elseif k == "right" then
		current.cursor = move_cursor(current.cursor, 1, 0, root)
	elseif (k == "space" or k == "return") then
		if not current.cursor then
			current.cursor = 1
		else
			click_on_slot(current.cursor)
		end
	elseif k == "s" then
		-- skip level
		init(current.level + 1)
	elseif k == "r" then
		-- restart level
		init(current.level)
	end
end


function love.displayrotated()
	last_rotation_time = love.timer.getTime()
end


function love.quit()
	if current.level > 0 then
		save_game({game, current})
	end
end
